#!/usr/bin/env python
# coding: utf-8
import unittest
import sys
from optparse import OptionParser
import logging
from copy import copy

from test.util import (prepare_test_environment, clear_test_environment,
                       GLOBAL, start_server, stop_server)
from weblib.watch import watch

# **********
# Grab Tests
# * pycurl transport
# * extensions
# **********
GRAB_TEST_LIST = (
    # Internal API
    'test.grab_api',
    'test.grab_transport',
    'test.response_class',
    'test.grab_debug',
    # Response processing
    'test.grab_xml_processing',
    'test.grab_response_body_processing',
    'test.grab_charset',
    # Network
    'test.grab_get_request',
    'test.grab_post_request',
    'test.grab_request',
    'test.grab_user_agent',
    'test.grab_cookies',
    'test.grab_url_processing',
    # Refactor
    'test.grab_proxy',
    'test.grab_upload_file',
    'test.grab_limit_option',
    'test.grab_charset_issue',
    'test.grab_pickle',
    # *** Extension sub-system
    # *** Extensions
    'test.ext_text',
    'test.ext_rex',
    'test.ext_lxml',
    'test.ext_form',
    'test.ext_doc',
    'test.ext_structured',
    # *** Pycurl Test
    'test.pycurl_cookie',
    # *** util.module
    'test.util_module',
    'test.util_log',
    # *** grab.export
    'test.util_config',
    'test.script_crawl',
    'test.script_start_project',
    'test.grab_error',
    'test.selector_deprecated',
    'test.grab_deprecated',
    'test.ext_pyquery',
    'test.tools_deprecated',
)

# ************
# Spider Tests
# ************

SPIDER_TEST_LIST = (
    'test.spider_task',
    'test.spider',
    'test.spider_proxy',
    'test.spider_queue',
    'test.spider_misc',
    'test.spider_meta',
    'test.spider_error',
    'test.spider_cache',
    'test.spider_data',
    'test.spider_stat',
    'test.spider_multiprocess',
)


def main():
    logging.basicConfig(level=logging.DEBUG)
    parser = OptionParser()
    parser.add_option('-t', '--test', help='Run only specified tests')
    parser.add_option('--test-grab', action='store_true',
                      default=False, help='Run tests for Grab::Spider')
    parser.add_option('--test-spider', action='store_true',
                      default=False, help='Run tests for Grab')
    parser.add_option('--test-all', action='store_true',
                      default=False,
                      help='Run tests for both Grab and Grab::Spider')
    parser.add_option('--backend-mongo', action='store_true',
                      default=False,
                      help='Run extra tests that depends on mongodb')
    parser.add_option('--backend-redis', action='store_true',
                      default=False,
                      help='Run extra tests that depends on redis')
    parser.add_option('--backend-mysql', action='store_true',
                      default=False,
                      help='Run extra tests that depends on mysql')
    parser.add_option('--backend-postgresql', action='store_true',
                      default=False,
                      help='Run extra tests that depends on postgresql')
    parser.add_option('--multiprocess', action='store_true', default=False,
                      help='Use multiprocess spider in spider tests')
    opts, args = parser.parse_args()

    if opts.backend_mongo:
        GLOBAL['backends'].append('mongo')

    if opts.backend_redis:
        GLOBAL['backends'].append('redis')

    if opts.backend_mysql:
        GLOBAL['backends'].append('mysql')

    if opts.backend_postgresql:
        GLOBAL['backends'].append('postgresql')

    prepare_test_environment()
    test_list = []

    if opts.test_all:
        test_list += GRAB_TEST_LIST
        test_list += SPIDER_TEST_LIST

    if opts.test_grab:
        test_list += GRAB_TEST_LIST

    if opts.test_spider:
        test_list += SPIDER_TEST_LIST

    if opts.test:
        test_list += [opts.test]

    GLOBAL['multiprocess'] = opts.multiprocess

    # Check tests integrity
    # Ensure that all test modules are imported correctly
    for path in test_list:
        __import__(path, None, None, ['foo'])

    loader = unittest.TestLoader()
    suite = unittest.TestSuite()
    for path in test_list:
        mod_suite = loader.loadTestsFromName(path)
        for some_suite in mod_suite:
            for test in some_suite:
                if (not hasattr(test, '_backend') or
                        test._backend in GLOBAL['backends']):
                    suite.addTest(test)

    runner = unittest.TextTestRunner()
    start_server()
    result = runner.run(suite)
    stop_server()
    clear_test_environment()
    if result.wasSuccessful():
        sys.exit(0)
    else:
        sys.exit(1)

if __name__ == '__main__':
    main()



# coding: utf-8
from test.util import build_grab, TMP_FILE
from test.util import (BaseGrabTestCase, TEST_SERVER_PORT,
                       EXTRA_PORT1, EXTRA_PORT2)
from test_server import TestServer
import six

ADDRESS = '127.0.0.1'
PROXY1 = '%s:%d' % (ADDRESS, TEST_SERVER_PORT)
PROXY2 = '%s:%d' % (ADDRESS, EXTRA_PORT1)
PROXY3 = '%s:%d' % (ADDRESS, EXTRA_PORT2)


class TestProxy(BaseGrabTestCase):
    def test_proxy_option(self):
        g = build_grab()

        g.setup(proxy=PROXY1, proxy_type='http', debug=True)
        self.server.response['get.data'] = '123'

        g.go('http://yandex.ru')
        self.assertEqual(b'123', g.response.body)
        self.assertEqual('yandex.ru', self.server.request['headers']['host'])

    def test_deprecated_setup_proxylist(self):
        g = build_grab()
        open(TMP_FILE, 'w').write(PROXY1)
        g.load_proxylist(TMP_FILE, 'text_file')
        self.server.response['get.data'] = '123'
        g.change_proxy()
        g.go('http://yandex.ru')
        self.assertEqual(b'123', g.response.body)
        self.assertEqual('yandex.ru', self.server.request['headers']['host'])

    def test_load_proxylist(self):
        content = '%s\n%s\n%s' % (PROXY1, PROXY2, PROXY3)
        open(TMP_FILE, 'w').write(content)

        # By default auto_change is True
        g = build_grab()
        g.load_proxylist(TMP_FILE, 'text_file')
        self.assertEqual(g.config['proxy_auto_change'], True)
        servers = set()
        for x in six.moves.range(10):
            g.go('http://yandex.ru')
            servers.add(g.config['proxy'])

        self.assertTrue(len(servers) > 1)

        # Disable auto_change
        # By default auto_init is True
        g = build_grab()
        g.load_proxylist(TMP_FILE, 'text_file', auto_change=False)
        self.assertEqual(g.config['proxy_auto_change'], False)
        servers = set()
        for x in six.moves.range(10):
            g.go('http://yandex.ru')
            servers.add(g.config['proxy'])
        self.assertEqual(len(servers), 1)

        # Disable auto_change
        # Disable auto_init
        # Proxylist will not be used by default
        g = build_grab()
        g.load_proxylist(TMP_FILE, 'text_file', auto_change=False,
                         auto_init=False)
        self.assertEqual(g.config['proxy_auto_change'], False)
        g.go(self.server.get_url())
        self.assertEqual(g.config['proxy'], None)

    def test_change_proxy(self):
        g = build_grab()
        g.change_proxy()
        self.assertEqual(g.config['proxy'], None)

        g = build_grab()
        with open(TMP_FILE, 'w') as out:
            for x in six.moves.range(10):
                out.write('server-%d:777\n' % x)
        g.load_proxylist(TMP_FILE, 'text_file', auto_init=False,
                         auto_change=False)
        self.assertEqual(g.config['proxy'], None)

        g.load_proxylist(TMP_FILE, 'text_file', auto_init=False,
                         auto_change=True)
        self.assertEqual(g.config['proxy'], None)

        g.load_proxylist(TMP_FILE, 'text_file', auto_init=True,
                         auto_change=False)
        self.assertTrue('server-' in g.config['proxy'])



from unittest import TestCase

from grab.util.module import import_string, ImportStringError


class UtilModuleTestCase(TestCase):
    def test_import_string(self):
        mod = import_string('test.case.util_module')
        self.assertTrue(hasattr(mod, 'UtilModuleTestCase'))

    def test_import_string_error(self):
        self.assertRaises(ImportStringError, import_string,
                          'test.case.zzzzzzzzzzzz')



from __future__ import absolute_import
from datetime import datetime, timedelta

from grab.spider.error import SpiderMisuseError
from grab.base import copy_config


class BaseTask(object):
    pass


class Task(BaseTask):
    """
    Task for spider.
    """

    def __init__(self, name=None, url=None, grab=None, grab_config=None,
                 priority=None, priority_is_custom=True,
                 network_try_count=0, task_try_count=1,
                 disable_cache=False, refresh_cache=False,
                 valid_status=[], use_proxylist=True,
                 cache_timeout=None, delay=0,
                 raw=False, callback=None,
                 fallback_name=None,
                 **kwargs):
        """
        Create `Task` object.

        If more than one of url, grab and grab_config options are non-empty
        then they processed in following order:
        * grab overwrite grab_config
        * grab_config overwrite url

        Args:
            :param name: name of the task. After successful network operation
                task's result will be passed to `task_<name>` method.
            :param url: URL of network document. Any task requires `url` or
                `grab` option to be specified.
            :param grab: configured `Grab` instance. You can use that option in
                case when `url` option is not enough. Do not forget to
                configure `url` option of `Grab` instance because in this case
                the `url` option of `Task` constructor will be overwritten
                with `grab.config['url']`.
            :param priority: - priority of the Task. Tasks with lower priority
                will be processed earlier. By default each new task is assigned
                with random priority from (80, 100) range.
            :param priority_is_custom: - internal flag which tells if that task
                priority was assigned manually or generated by spider according
                to priority generation rules.
            :param network_try_count: you'll probably will not need to use it.
                It is used internally to control how many times this task was
                restarted due to network errors. The `Spider` instance has
                `network_try_limit` option. When `network_try_count` attribute
                of the task exceeds the `network_try_limit` attribute then
                processing of the task is abandoned.
            :param task_try_count: the as `network_try_count` but it increased
                only then you use `clone` method. Also you can set it manually.
                It is useful if you want to restart the task after it was
                cancelled due to multiple network errors. As you might guessed
                there is `task_try_limit` option in `Spider` instance. Both
                options `network_try_count` and `network_try_limit` guarantee
                you that you'll not get infinite loop of restarting some task.
            :param disable_cache: if `True` disable cache subsystem.
                The document will be fetched from the Network and it will not
                be saved to cache.
            :param refresh_cache: if `True` the document will be fetched from
                the Network and saved to cache.
            :param valid_status: extra status codes which counts as valid
            :param use_proxylist: it means to use proxylist which was
                configured via `setup_proxylist` method of spider
            :param cache_timeout: maximum age (in seconds) of cache record to
                be valid
            :param delay: if specified tells the spider to schedule the task
                and execute    it after `delay` seconds
            :param raw: if `raw` is True then the network response is
                forwarding to the corresponding handler without any check of
                HTTP status code of network error, if `raw` is False (by
                default) then failed response is putting back to task queue or
                if tries limit is reached then the processing of this  request
                is finished.
            :param callback: if you pass some function in `callback` option
                then the network response will be passed to this callback and
                the usual 'task_*' handler will be ignored and no error will be
                raised if such 'task_*' handler does not exist.
            :param fallback_name: the name of method that is called when spider
                gives up to do the task (due to multiple network errors)

            Any non-standard named arguments passed to `Task` constructor will
            be saved as attributes of the object. You can get their values
            later as attributes or with `get` method which allows to use
            default value if attribute does not exist.
        """

        if name == 'generator':
            # The name "generator" is restricted because
            # `task_generator` handler could not be created because
            # this name is already used for special method which
            # generates new tasks
            raise SpiderMisuseError('Task name could not be "generator"')

        self.name = name

        if url is None and grab is None and grab_config is None:
            raise SpiderMisuseError('Either url, grab or grab_config argument '
                                    'of Task constructor should not be None')

        if url is not None and grab is not None:
            raise SpiderMisuseError('Options url and grab could not be used '
                                    'together')

        if url is not None and grab_config is not None:
            raise SpiderMisuseError('Options url and grab_config could not be '
                                    'used together')

        if grab is not None and grab_config is not None:
            raise SpiderMisuseError(
                'Options grab and grab_config could not be used together')

        if grab:
            self.setup_grab_config(grab.dump_config())
        elif grab_config:
            self.setup_grab_config(grab_config)
        else:
            self.grab_config = None
            self.url = url

        self.process_delay_option(delay)

        self.fallback_name = fallback_name
        self.priority_is_custom = priority_is_custom
        self.priority = priority
        self.network_try_count = network_try_count
        self.task_try_count = task_try_count
        self.disable_cache = disable_cache
        self.refresh_cache = refresh_cache
        self.valid_status = valid_status
        self.use_proxylist = use_proxylist
        self.cache_timeout = cache_timeout
        self.raw = raw
        self.origin_task_generator = None
        self.callback = callback
        self.coroutines_stack = []
        for key, value in kwargs.items():
            setattr(self, key, value)

    def get(self, key, default=None):
        """
        Return value of attribute or None if such attribute
        does not exist.
        """
        return getattr(self, key, default)

    def process_delay_option(self, delay):
        if delay:
            self.schedule_time = datetime.utcnow() + timedelta(seconds=delay)
            self.original_delay = delay
        else:
            self.schedule_time = None
            self.original_delay = None

    def setup_grab_config(self, grab_config):
        self.grab_config = copy_config(grab_config)
        self.url = grab_config['url']

    def clone(self, **kwargs):
        """
        Clone Task instance.

        Reset network_try_count, increase task_try_count.
        """

        # First, create exact copy of the current Task object
        attr_copy = self.__dict__.copy()
        if attr_copy.get('grab_config') is not None:
            del attr_copy['url']
        task = Task(**attr_copy)

        # Reset some task properties if they have not
        # been set explicitly in kwargs
        if 'network_try_count' not in kwargs:
            task.network_try_count = 0
        if 'task_try_count' not in kwargs:
            task.task_try_count = self.task_try_count + 1
        if 'refresh_cache' not in kwargs:
            task.refresh_cache = False
        if 'disable_cache' not in kwargs:
            task.disable_cache = False

        if kwargs.get('url') is not None and kwargs.get('grab') is not None:
            raise SpiderMisuseError('Options url and grab could not be '
                                    'used together')

        if (kwargs.get('url') is not None and
                kwargs.get('grab_config') is not None):
            raise SpiderMisuseError('Options url and grab_config could not '
                                    'be used together')

        if (kwargs.get('grab') is not None and
                kwargs.get('grab_config') is not None):
            raise SpiderMisuseError('Options grab and grab_config could not '
                                    'be used together')

        if kwargs.get('grab'):
            task.setup_grab_config(kwargs['grab'].dump_config())
            del kwargs['grab']
        elif kwargs.get('grab_config'):
            task.setup_grab_config(kwargs['grab_config'])
            del kwargs['grab_config']
        elif kwargs.get('url'):
            task.url = kwargs['url']
            if task.grab_config:
                task.grab_config['url'] = kwargs['url']
            del kwargs['url']

        for key, value in kwargs.items():
            setattr(task, key, value)

        # WTF?
        # The `Task` object can't has `delay` attribute
        # I think in next line the `process_delay_option` method
        # always gets None as input argument
        task.process_delay_option(task.get('delay', None))

        return task

    def __repr__(self):
        return '<Task: %s>' % self.url

    def __lt__(self, other):
        return self.priority < other.priority

    def __eq__(self, other):
        if not self.priority or not other.priority:
            return True
        else:
            return self.priority == other.priority

    def get_fallback_handler(self, spider):
        if self.fallback_name:
            return getattr(spider, self.fallback_name)
        elif self.name:
            fb_name = 'task_%s_fallback' % self.name
            if hasattr(spider, fb_name):
                return getattr(spider, fb_name)
        else:
            return None


def inline_task(f):
    def wrap(self, grab, task):
        origin_task_generator = f(self, grab, task)
        self.handler_for_inline_task(None, origin_task_generator)
    return wrap



# coding: utf-8
import six

from test.util import build_grab
from test.util import BaseGrabTestCase


class LXMLExtensionTest(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_dash_issue(self):
        HTML = '<strong>&#151;</strong>'
        self.server.response['get.data'] = HTML
        g = build_grab()
        g.go(self.server.get_url())

        # By default &#[128-160]; are fixed
        self.assertFalse(g.xpath_one('//strong/text()') == six.unichr(151))
        self.assertTrue(g.xpath_one('//strong/text()') == six.unichr(8212))

        # disable fix-behaviour
        g.setup(fix_special_entities=False)
        g.go(self.server.get_url())

        # By default &#[128-160]; are fixed
        self.assertTrue(g.xpath_one('//strong/text()') == six.unichr(151))
        self.assertFalse(g.xpath_one('//strong/text()') == six.unichr(8212))

        # Explicitly use unicode_body func
        g = build_grab()
        g.go(self.server.get_url())
        print(':::', g.response.unicode_body())
        self.assertTrue('&#8212;' in g.response.unicode_body())



import six

import grab.spider.base
from grab import Grab
from grab.spider import Spider, Task, SpiderMisuseError, NoTaskHandler
from grab.spider import inline_task
from test.util import (BaseGrabTestCase, build_grab, build_spider,
                       multiprocess_mode)
from grab.spider.error import SpiderError


class SimpleSpider(Spider):
    base_url = 'http://google.com'

    def task_baz(self, grab, task):
        self.SAVED_ITEM = grab.response.body


class TestSpider(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_task_priority(self):
        # Automatic random priority
        grab.spider.base.RANDOM_TASK_PRIORITY_RANGE = (10, 20)
        bot = build_spider(SimpleSpider, priority_mode='random')
        bot.setup_queue()
        task = Task('baz', url='xxx')
        self.assertEqual(task.priority, None)
        bot.add_task(task)
        self.assertTrue(10 <= task.priority <= 20)

        # Automatic constant priority
        grab.spider.base.DEFAULT_TASK_PRIORITY = 33
        bot = build_spider(SimpleSpider, priority_mode='const')
        bot.setup_queue()
        task = Task('baz', url='xxx')
        self.assertEqual(task.priority, None)
        bot.add_task(task)
        self.assertEqual(33, task.priority)

        # Automatic priority does not override explictily setted priority
        grab.spider.base.DEFAULT_TASK_PRIORITY = 33
        bot = build_spider(SimpleSpider, priority_mode='const')
        bot.setup_queue()
        task = Task('baz', url='xxx', priority=1)
        self.assertEqual(1, task.priority)
        bot.add_task(task)
        self.assertEqual(1, task.priority)

        self.assertRaises(SpiderMisuseError,
                          lambda: SimpleSpider(priority_mode='foo'))

    def test_task_url(self):
        bot = build_spider(SimpleSpider, )
        bot.setup_queue()
        task = Task('baz', url='xxx')
        self.assertEqual('xxx', task.url)
        bot.add_task(task)
        self.assertEqual('http://google.com/xxx', task.url)
        self.assertEqual(None, task.grab_config)

        g = Grab(url='yyy')
        task = Task('baz', grab=g)
        bot.add_task(task)
        self.assertEqual('http://google.com/yyy', task.url)
        self.assertEqual('http://google.com/yyy', task.grab_config['url'])

    def test_task_clone(self):
        bot = build_spider(SimpleSpider, )
        bot.setup_queue()

        task = Task('baz', url='xxx')
        bot.add_task(task.clone())

        # Pass grab to clone
        task = Task('baz', url='xxx')
        g = Grab()
        g.setup(url='zzz')
        bot.add_task(task.clone(grab=g))

        # Pass grab_config to clone
        task = Task('baz', url='xxx')
        g = Grab()
        g.setup(url='zzz')
        bot.add_task(task.clone(grab_config=g.config))

    def test_task_clone_with_url_param(self):
        task = Task('baz', url='xxx')
        task.clone(url='http://yandex.ru/')

    def test_task_useragent(self):
        bot = build_spider(SimpleSpider, )
        bot.setup_queue()

        g = Grab()
        g.setup(url=self.server.get_url())
        g.setup(user_agent='Foo')

        task = Task('baz', grab=g)
        bot.add_task(task.clone())
        bot.run()
        self.assertEqual(self.server.request['headers']['User-Agent'], 'Foo')

    def test_task_nohandler_error(self):
        class TestSpider(Spider):
            pass

        bot = build_spider(TestSpider, )
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        self.assertRaises(NoTaskHandler, bot.run)

    def test_task_raw(self):
        class TestSpider(Spider):
            def task_page(self, grab, task):
                self.stat.collect('codes', grab.response.code)

        self.server.response['code'] = 502

        bot = build_spider(TestSpider, network_try_limit=1)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.run()
        self.assertEqual(0, len(bot.stat.collections['codes']))

        bot = build_spider(TestSpider, network_try_limit=1)
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url(), raw=True))
        bot.add_task(Task('page', url=self.server.get_url(), raw=True))
        bot.run()
        self.assertEqual(2, len(bot.stat.collections['codes']))

    @multiprocess_mode(False)
    def test_task_callback(self):
        class TestSpider(Spider):
            def task_page(self, grab, task):
                self.meta['tokens'].append('0_handler')

        class FuncWithState(object):
            def __init__(self, tokens):
                self.tokens = tokens

            def __call__(self, grab, task):
                self.tokens.append('1_func')

        tokens = []
        func = FuncWithState(tokens)

        bot = build_spider(TestSpider, )
        bot.meta['tokens'] = tokens
        bot.setup_queue()
        # classic handler
        bot.add_task(Task('page', url=self.server.get_url()))
        # callback option overried classic handler
        bot.add_task(Task('page', url=self.server.get_url(), callback=func))
        # callback and null task name
        bot.add_task(Task(name=None, url=self.server.get_url(), callback=func))
        # callback and default task name
        bot.add_task(Task(url=self.server.get_url(), callback=func))
        bot.run()
        self.assertEqual(['0_handler', '1_func', '1_func', '1_func'],
                         sorted(tokens))

    @multiprocess_mode(False)
    def test_inline_task(self):

        def callback(self):
            self.write(self.request.uri)
            self.finish()

        self.server.response['get.callback'] = callback

        server = self.server

        class TestSpider(Spider):
            def add_response(self, grab):
                self.stat.collect('responses', grab.doc.unicode_body())

            def task_generator(self):
                url = server.get_url('/?foo=start')
                yield Task('inline', url=url)

            def subroutine_task(self, grab):

                for x in six.moves.range(2):
                    url = server.get_url('/?foo=subtask%s' % x)
                    grab.setup(url=url)
                    grab = yield Task(grab=grab)
                    self.add_response(grab)
                    self.stat.collect('calls', 'subinline%s' % x)

            @inline_task
            def task_inline(self, grab, task):
                self.add_response(grab)
                self.stat.collect('calls', 'generator')

                for x in six.moves.range(3):
                    url = server.get_url('/?foo=%s' % x)
                    grab.setup(url=url)
                    grab = yield Task(grab=grab)

                    self.add_response(grab)
                    self.stat.collect('calls', 'inline%s' % x)

                    grab = yield self.subroutine_task(grab)
                    # In this case the grab body will be the same
                    # as is in subroutine task:  /?foo=subtask1
                    self.add_response(grab)

                url = server.get_url('/?foo=yield')
                self.add_task(Task('yield', url=url))

            def task_yield(self, grab, task):
                self.add_response(grab)
                self.stat.collect('calls', 'yield')

                url = server.get_url('/?foo=end')
                yield Task('end', url=url)

            def task_end(self, grab, task):
                self.add_response(grab)
                self.stat.collect('calls', 'end')

        bot = build_spider(TestSpider, )
        bot.run()

        self.assertEqual(['/?foo=start',
                          '/?foo=0',
                          '/?foo=subtask0', '/?foo=subtask1', '/?foo=subtask1',
                          '/?foo=1',
                          '/?foo=subtask0', '/?foo=subtask1', '/?foo=subtask1',
                          '/?foo=2',
                          '/?foo=subtask0', '/?foo=subtask1', '/?foo=subtask1',
                          '/?foo=yield', '/?foo=end'],
                         bot.stat.collections['responses'])
        self.assertEqual(['generator',
                          'inline0',
                          'subinline0', 'subinline1',
                          'inline1',
                          'subinline0', 'subinline1',
                          'inline2',
                          'subinline0', 'subinline1',
                          'yield', 'end'],
                         bot.stat.collections['calls'])

    def test_task_url_and_grab_options(self):
        class TestSpider(Spider):
            def setup(self):
                self.done = False

            def task_page(self, grab, task):
                self.done = True

        bot = build_spider(TestSpider, )
        bot.setup_queue()
        g = Grab()
        g.setup(url=self.server.get_url())
        self.assertRaises(SpiderMisuseError, Task,
                          'page', grab=g, url=self.server.get_url())

    def test_task_invalid_name(self):
        self.assertRaises(SpiderMisuseError, Task,
                          'generator', url='http://ya.ru/')

    def test_task_constructor_invalid_args(self):
        # no url, no grab, no grab_config
        self.assertRaises(SpiderMisuseError, Task, 'foo')
        # both url and grab_config
        self.assertRaises(SpiderMisuseError, Task, 'foo',
                          url=1, grab_config=1)
        # both grab and grab_config
        self.assertRaises(SpiderMisuseError, Task, 'foo',
                          grab=1, grab_config=1)

    def test_task_clone_invalid_args(self):
        task = Task('foo', url='http://ya.ru/')
        # both url and grab
        self.assertRaises(SpiderMisuseError, task.clone,
                          url=1, grab=1)
        # both url and grab_config
        self.assertRaises(SpiderMisuseError, task.clone,
                          url=1, grab_config=1)
        # both grab_config and grab
        self.assertRaises(SpiderMisuseError, task.clone,
                          grab=1, grab_config=1)

    def test_task_clone_grab_config_and_url(self):
        g = build_grab()
        g.setup(url='http://foo.com/')
        task = Task('foo', grab=g)
        task2 = task.clone(url='http://bar.com/')
        self.assertEqual(task2.url, 'http://bar.com/')
        self.assertEqual(task2.grab_config['url'], 'http://bar.com/')

    def test_task_clone_kwargs(self):
        g = build_grab()
        g.setup(url='http://foo.com/')
        task = Task('foo', grab=g, cache_timeout=1)
        task2 = task.clone(cache_timeout=2)
        self.assertEqual(2, task2.cache_timeout)

    def test_task_comparison(self):
        t1 = Task('foo', url='http://foo.com/', priority=1)
        t2 = Task('foo', url='http://foo.com/', priority=2)
        t3 = Task('foo', url='http://foo.com/')
        # If both tasks have priorities then task are
        # compared by their priorities
        self.assertTrue(t1 < t2)
        # If any of compared tasks does not have priority
        # than tasks are equal
        self.assertTrue(t1 == t3)
        self.assertTrue(t3 == t3)

    def test_task_get_fallback_handler(self):
        class TestSpider(Spider):
            def zz(self, task):
                pass

            def task_bar_fallback(self, task):
                pass


        t1 = Task('foo', url='http://foo.com/', fallback_name='zz')
        t2 = Task('bar', url='http://foo.com/')
        t3 = Task(url='http://foo.com/')

        bot = build_spider(TestSpider, )

        self.assertEqual(t1.get_fallback_handler(bot), bot.zz)
        self.assertEqual(t2.get_fallback_handler(bot), bot.task_bar_fallback)
        self.assertEqual(t3.get_fallback_handler(bot), None)

    def test_update_grab_instance(self):
        class TestSpider(Spider):
            def update_grab_instance(self, grab):
                grab.setup(timeout=77)

            def task_generator(self):
                yield Task('page', url=self.meta['server'].get_url())
                yield Task('page', grab=Grab(url=self.meta['server'].get_url(),
                                             timeout=1))

            def task_page(self, grab, task):
                self.stat.collect('points', grab.config['timeout'])

        bot = build_spider(TestSpider, meta={'server': self.server})
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.add_task(Task('page', grab=Grab(url=self.server.get_url(),
                                            timeout=1)))
        bot.run()
        self.assertEqual(set([77]), set(bot.stat.collections['points']))

    def test_create_grab_instance(self):
        class TestSpider(Spider):
            def create_grab_instance(self, **kwargs):
                grab = super(TestSpider, self).create_grab_instance(**kwargs)
                grab.setup(timeout=77)
                return grab

            def task_generator(self):
                yield Task('page', url=self.meta['server'].get_url())
                yield Task('page', grab=Grab(url=self.meta['server'].get_url(),
                                             timeout=76))

            def task_page(self, grab, task):
                self.stat.collect('points', grab.config['timeout'])

        bot = build_spider(TestSpider, meta={'server': self.server})
        bot.setup_queue()
        bot.add_task(Task('page', url=self.server.get_url()))
        bot.add_task(Task('page', grab=Grab(url=self.server.get_url(),
                                            timeout=75)))
        bot.run()
        self.assertEqual(set([77, 76, 75]),
                         set(bot.stat.collections['points']))

    def test_add_task_invalid_url_no_error(self):
        class TestSpider(Spider):
            pass

        bot = build_spider(TestSpider, )
        bot.setup_queue()
        bot.add_task(Task('page', url='zz://zz'))
        self.assertEqual(0, bot.taskq.size())
        bot.add_task(Task('page', url='zz://zz'), raise_error=False)
        self.assertEqual(0, bot.taskq.size())
        bot.add_task(Task('page', url='http://example.com/'))
        self.assertEqual(1, bot.taskq.size())

    def test_add_task_invalid_url_raise_error(self):
        class TestSpider(Spider):
            pass

        bot = build_spider(TestSpider, )
        bot.setup_queue()
        self.assertRaises(SpiderError, bot.add_task,
                          Task('page', url='zz://zz'), raise_error=True)
        self.assertEqual(0, bot.taskq.size())
        bot.add_task(Task('page', url='http://example.com/'))
        self.assertEqual(1, bot.taskq.size())

    def test_multiple_internal_worker_error(self):
        class TestSpider(Spider):
            def process_network_result_with_handler_mp(*args, **kwargs):
                1/0

            def task_page(self):
                pass

        bot = build_spider(TestSpider, )
        bot.setup_queue()
        for x in range(5):
            bot.add_task(Task('page', url='http://ya.ru/'))
        bot.run()
        self.assertTrue(1 < bot.stat.counters['parser-process-restore'])



# coding: utf-8
from test.util import build_grab, get_temp_file
from test.util import BaseGrabTestCase

from grab import Grab
from grab.proxylist import ProxyList

DEFAULT_PLIST_DATA = \
    '1.1.1.1:8080\n'\
    '1.1.1.2:8080\n'


class GrabProxyTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def generate_plist_file(self, data=DEFAULT_PLIST_DATA):
        path = get_temp_file()
        with open(path, 'w') as out:
            out.write(data)
        return path

    def test_no_proxy_list(self):
        g = build_grab()
        self.assertEqual(0, g.proxylist.size())


class ProxyListTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def generate_plist_file(self, data=DEFAULT_PLIST_DATA):
        path = get_temp_file()
        with open(path, 'w') as out:
            out.write(data)
        return path

    def test_basic(self):
        pl = ProxyList()
        self.assertEqual(0, pl.size())


    def test_file_proxy_source(self):
        pl = ProxyList()
        path = self.generate_plist_file()
        pl.load_file(path)
        self.assertEqual(2, pl.size())

    def test_web_proxy_source(self):
        pl = ProxyList()
        self.server.response['data'] = DEFAULT_PLIST_DATA
        pl.load_url(self.server.get_url())
        self.assertEqual(2, pl.size())

    def test_get_next_proxy(self):
        pl = ProxyList()
        path = self.generate_plist_file('foo:1\nbar:1')
        pl.load_file(path)
        self.assertEqual(pl.get_next_proxy().host, 'foo')
        self.assertEqual(pl.get_next_proxy().host, 'bar')
        self.assertEqual(pl.get_next_proxy().host, 'foo')
        pl.load_file(path)
        self.assertEqual(pl.get_next_proxy().host, 'foo')



# coding: utf-8
from unittest import TestCase

from test.util import build_grab

HTML = b"""
<html>
    <body>
        <h1>test</h1>
    </body>
</html>
"""


class DocExtensionTest(TestCase):
    def setUp(self):
        # Create fake grab instance with fake response
        self.g = build_grab(document_body=HTML)

    def test_extension_in_general(self):
        self.assertTrue(self.g.doc)

    def test_select_method(self):
        self.assertEqual('test', self.g.doc.select('//h1').text())



# coding: utf-8
try:
    from urllib import quote
except ImportError:
    from urllib.parse import quote

from grab import GrabMisuseError
from test.util import build_grab
from test.util import BaseGrabTestCase
try:
    from urlparse import parse_qsl
except ImportError:
    from urllib.parse import parse_qsl


class TestPostFeature(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_post(self):
        g = build_grab(url=self.server.get_url(), debug_post=True)

        # Provide POST data in dict
        g.setup(post={'foo': 'bar'})
        g.request()
        self.assertEqual(self.server.request['data'], b'foo=bar')

        # Provide POST data in tuple
        g.setup(post=(('foo', 'TUPLE'),))
        g.request()
        self.assertEqual(self.server.request['data'], b'foo=TUPLE')

        # Provide POST data in list
        g.setup(post=[('foo', 'LIST')])
        g.request()
        self.assertEqual(self.server.request['data'], b'foo=LIST')

        # Order of elements should not be changed (1)
        g.setup(post=[('foo', 'LIST'), ('bar', 'BAR')])
        g.request()
        self.assertEqual(self.server.request['data'], b'foo=LIST&bar=BAR')

        # Order of elements should not be changed (2)
        g.setup(post=[('bar', 'BAR'), ('foo', 'LIST')])
        g.request()
        self.assertEqual(self.server.request['data'], b'bar=BAR&foo=LIST')

        # Provide POST data in byte-string
        g.setup(post='Hello world!')
        g.request()
        self.assertEqual(self.server.request['data'], b'Hello world!')

        # Provide POST data in unicode-string
        g.setup(post=u'Hello world!')
        g.request()
        self.assertEqual(self.server.request['data'], b'Hello world!')

        # Provide POST data in non-ascii unicode-string
        g.setup(post=u'Привет, мир!')
        g.request()
        self.assertEqual(self.server.request['data'],
                         u'Привет, мир!'.encode('utf-8'))

        # Two values with one key
        g.setup(post=(('foo', 'bar'), ('foo', 'baz')))
        g.request()
        self.assertEqual(self.server.request['data'], b'foo=bar&foo=baz')

    def assertEqualQueryString(self, qs1, qs2):
        args1 = set([(x, y[0]) for x, y in parse_qsl(qs1)])
        args2 = set([(x, y[0]) for x, y in parse_qsl(qs2)])
        self.assertEqual(args1, args2)

    def test_multipart_post(self):
        g = build_grab(url=self.server.get_url(), debug_post=True)
        # Dict
        g.setup(multipart_post={'foo': 'bar'})
        g.request()
        self.assertTrue(b'name="foo"' in self.server.request['data'])

        # Few values with non-ascii data
        # TODO: understand and fix
        # AssertionError: 'foo=bar&gaz=%D0%94%D0%B5%D0%BB%'\
        #                 'D1%8C%D1%84%D0%B8%D0%BD&abc=' !=
        #                 'foo=bar&gaz=\xd0\x94\xd0\xb5\xd0'\
        #                 '\xbb\xd1\x8c\xd1\x84\xd0\xb8\xd0\xbd&abc='
        # g.setup(post=({'foo': 'bar', 'gaz': u'Дельфин', 'abc': None}))
        # g.request()
        # self.assertEqual(self.server.request['data'],
        #                   'foo=bar&gaz=Дельфин&abc=')

        # Multipart data could not be string
        g.setup(multipart_post='asdf')
        self.assertRaises(GrabMisuseError, lambda: g.request())

        # tuple with one pair
        g.setup(multipart_post=(('foo', 'bar'),))
        g.request()
        self.assertTrue(b'name="foo"' in self.server.request['data'])

        # tuple with two pairs
        g.setup(multipart_post=(('foo', 'bar'), ('foo', 'baz')))
        g.request()
        self.assertTrue(b'name="foo"' in self.server.request['data'])

    def test_unicode_post(self):
        # By default, unicode post should be converted into utf-8
        g = build_grab()
        data = u'фыва'
        g.setup(post=data, url=self.server.get_url())
        g.request()
        self.assertEqual(self.server.request['data'], data.encode('utf-8'))

        # Now try cp1251 with charset option
        self.server.request['charset'] = 'cp1251'
        g = build_grab()
        data = u'фыва'
        g.setup(post=data, url=self.server.get_url(),
                charset='cp1251', debug=True)
        g.request()
        self.assertEqual(self.server.request['data'], data.encode('cp1251'))

        # Now try dict with unicode value & charset option
        self.server.request['charset'] = 'cp1251'
        g = build_grab()
        data = u'фыва'
        g.setup(post={'foo': data}, url=self.server.get_url(),
                charset='cp1251', debug=True)
        g.request()
        test = 'foo=%s' % quote(data.encode('cp1251'))
        test = test.encode('utf-8')  # py3 hack
        self.assertEqual(self.server.request['data'], test)

    def test_put(self):
        g = build_grab()
        g.setup(post=b'abc', url=self.server.get_url(),
                method='put', debug=True)
        self.server.request['debug'] = True
        g.request()
        self.assertEqual(self.server.request['method'], 'PUT')
        self.assertEqual(self.server.request['headers']['content-length'], '3')

    def test_patch(self):
        g = build_grab()
        g.setup(post=b'abc', url=self.server.get_url(), method='patch')
        g.request()
        self.assertEqual(self.server.request['method'], 'PATCH')
        self.assertEqual(self.server.request['headers']['content-length'], '3')

    def test_empty_post(self):
        g = build_grab()
        g.setup(method='post')
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['method'], 'POST')
        self.assertEqual(self.server.request['data'], b'')
        self.assertEqual(self.server.request['headers']['content-length'], '0')

        g.go(self.server.get_url(), post='DATA')
        self.assertEqual(self.server.request['headers']['content-length'], '4')



from grab.spider import Spider, Task
from grab.script import crawl
from grab.util.module import SPIDER_REGISTRY

from test.util import TMP_DIR
from test.util import BaseGrabTestCase


class TestSpider(Spider):
    url = None
    points = []

    def prepare(self):
        del self.points[:]

    def task_generator(self):
        yield Task('page', url=self.url)

    def task_page(self, grab, task):
        self.points.append(grab.response.body)



class FailSpider(Spider):
    def task_generator(self):
        yield Task('page', url=self.url)

    def task_page(self, grab, task):
        1/0


class ScriptCrawlTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_crawl(self):
        TestSpider.url = self.server.get_url()
        self.server.response['data'] = b'1'
        SPIDER_REGISTRY.clear()
        crawl.main('test_spider', settings_module='test.files.crawl_settings',
                   disable_report=True)
        self.assertEquals(TestSpider.points, [b'1'])

    def test_crawl_save_lists(self):
        FailSpider.url = self.server.get_url()
        self.server.response['data'] = b'1'
        SPIDER_REGISTRY.clear()
        crawl.main('fail_spider', settings_module='test.files.crawl_settings')



# coding: utf-8
"""
This test fails in py3.3 environment because `grab.response.body`
contains <str>, but it should contains <bytes>
"""
from test.util import build_grab
from test.util import BaseGrabTestCase


class GrabCharsetDetectionTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_document_charset_option(self):
        g = build_grab()
        self.server.response['get.data'] = b'foo'
        g.go(self.server.get_url())
        self.assertEqual(b'foo', g.response.body)

        g = build_grab()
        self.server.response['get.data'] = u'фуу'.encode('utf-8')
        g.go(self.server.get_url())
        self.assertEqual(u'фуу'.encode('utf-8'), g.response.body)

        print(g.response.head)
        self.assertEqual(g.response.charset, 'utf-8')

        g = build_grab(document_charset='cp1251')
        self.server.response['get.data'] = u'фуу'.encode('cp1251')
        g.go(self.server.get_url())
        self.assertEqual(u'фуу'.encode('cp1251'), g.response.body)
        self.assertEqual(g.response.charset, 'cp1251')

    def test_document_charset_lowercase(self):
        self.server.response['charset'] = 'UTF-8'
        g = build_grab()
        g.go(self.server.get_url())
        self.assertEquals('utf-8', g.doc.charset)



# coding: utf-8
from grab import GrabMisuseError, GrabError
from grab.error import GrabTooManyRedirectsError
from grab.base import reset_request_counter
from test.util import build_grab
from test.util import BaseGrabTestCase
import six
import tempfile
import os


class GrabApiTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_incorrect_option_name(self):
        g = build_grab()
        self.assertRaises(GrabMisuseError, g.setup,
                          save_the_word=True)

    # Ghost test was disabled because of strange error
    # that appear when multiple Ghost instances are created
    def test_clone(self):
        g = build_grab()
        self.server.response['get.data'] = 'Moon'
        g.go(self.server.get_url())
        self.assertTrue(b'Moon' in g.response.body)
        self.server.response['post.data'] = 'Foo'
        g2 = g.clone(method='post')
        g2.go(self.server.get_url())
        self.assertTrue(b'Foo' in g2.response.body)

    def test_empty_clone(self):
        g = build_grab()
        g.clone()

    def test_adopt(self):
        g = build_grab()
        self.server.response['get.data'] = 'Moon'
        g.go(self.server.get_url())
        g2 = build_grab()
        self.assertEqual(g2.config['url'], None)
        g2.adopt(g)
        self.assertTrue(b'Moon' in g2.response.body)
        self.assertEqual(g2.config['url'], self.server.get_url())

    def test_empty_adopt(self):
        g = build_grab()
        g2 = build_grab()
        g2.adopt(g)

    def test_default_content_for_fake_response(self):
        content = b'<strong>test</strong>'
        g = build_grab(document_body=content)
        self.assertEqual(g.response.body, content)

    def test_inheritance(self):
        from grab import Grab

        class SimpleExtension(object):
            data = {'counter': 0}

            @classmethod
            def get_data(cls):
                return cls.data

        class CustomGrab(Grab, SimpleExtension):
            pass

        SimpleExtension.get_data()['counter'] = 0
        CustomGrab()
        # self.assertEqual(SimpleExtension.get_data()['counter'], 1)

        class VeryCustomGrab(CustomGrab):
            pass

        SimpleExtension.get_data()['counter'] = 0
        VeryCustomGrab()
        # self.assertEqual(SimpleExtension.get_data()['counter'], 1)

    def test_request_counter(self):
        import threading

        reset_request_counter()
        g = build_grab()
        g.go(self.server.get_url())
        self.assertEqual(g.request_counter, 1)

        g.go(self.server.get_url())
        self.assertEqual(g.request_counter, 2)

        def func():
            g = build_grab()
            g.go(self.server.get_url())

        # Make 10 requests in concurrent threads
        threads = []
        for x in six.moves.range(10):
            th = threading.Thread(target=func)
            threads.append(th)
            th.start()
        for th in threads:
            th.join()

        g.go(self.server.get_url())
        self.assertEqual(g.request_counter, 13)

    def test_download(self):
        fd, path = tempfile.mkstemp()
        g = build_grab()
        self.server.response['get.data'] = 'FOO'
        length = g.download(self.server.get_url(), path)
        self.assertEqual(3, length)
        os.unlink(path)

    def test_follow_refresh_option(self):
        def handler():
            response = "<meta http-equiv='refresh' content='0;url= %s'>" % \
                       self.server.get_url()
            yield response.encode('ascii')
            yield response.encode('ascii')
            yield response.encode('ascii')
            yield b'OK'

        self.server.response['data'] = handler()
        g = build_grab(follow_refresh=True)
        g.go(self.server.get_url())
        self.assertEqual(g.response.body, b'OK')
        self.server.response['data'] = handler()
        g.setup(redirect_limit=1)
        self.assertRaises(GrabTooManyRedirectsError, g.go,
                          self.server.get_url())

    def test_make_url_absolute(self):
        g = build_grab()
        self.server.response['get.data'] = '<base href="http://foo/bar/">'
        g.go(self.server.get_url())
        absolute_url = g.make_url_absolute('/foobar', resolve_base=True)
        self.assertEqual(absolute_url, 'http://foo/foobar')
        g = build_grab()
        absolute_url = g.make_url_absolute('/foobar')
        self.assertEqual(absolute_url, '/foobar')

    def test_error_request(self):
        g = build_grab()
        g.setup(post={'foo': 'bar'})

        self.assertRaises(GrabError, g.go,
                          url='Could not resolve host address')
        self.assertEqual(g.config['post'], None)
        self.assertEqual(g.config['multipart_post'], None)
        self.assertEqual(g.config['method'], None)
        self.assertEqual(g.config['body_storage_filename'], None)
        self.assertEqual(g.config['refresh_redirect_count'], 0)

    def test_setup_document(self):
        data = b'''
        <h1>test</h1>
        '''
        g = build_grab(data)
        self.assertTrue(b'test' in g.doc.body)

    def test_setup_document_invalid_input(self):
        data = u'''
        <h1>test</h1>
        '''
        self.assertRaises(GrabMisuseError, build_grab, data)



# coding: utf-8
from grab import UploadContent, UploadFile
import tempfile
import os
import pycurl

from test.util import build_grab
from test.util import BaseGrabTestCase


class TestUploadContent(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def prepare_form_grab(self):
        url = self.server.get_url()
        html = ('''<form action="%s" method="post" enctype="multipart/form-data">
            <input type="file" name="image">
        </form>''' % url).encode('ascii')
        g = build_grab(html, charset='utf-8')
        return g

    # *******************
    # UploadContent Tests
    # *******************

    def test_upload_content_filename(self):
        g = self.prepare_form_grab()
        data = b'foo'
        upload_data = UploadContent(data, filename='avatar.jpg')
        g.doc.set_input('image', upload_data)
        g.doc.submit(make_request=False)
        post = dict(g.config['multipart_post'])
        self.assertTrue(isinstance(post['image'], UploadContent))

        g.doc.submit()
        self.assertEqual(data,
                         self.server.request['files']['image'][0]['body'])
        self.assertEqual('avatar.jpg',
                         self.server.request['files']['image'][0]['filename'])
        self.assertEqual(
            'image/jpeg',
            self.server.request['files']['image'][0]['content_type'])

    def test_upload_content_random_filename(self):
        g = self.prepare_form_grab()
        data = b'foo'
        upload_data = UploadContent(data)
        g.doc.set_input('image', upload_data)
        g.doc.submit(make_request=False)
        post = dict(g.config['multipart_post'])
        self.assertTrue(isinstance(post['image'], UploadContent))

        g.doc.submit()
        self.assertEqual(data,
                         self.server.request['files']['image'][0]['body'])
        self.assertTrue(
            10, len(self.server.request['files']['image'][0]['filename']))
        self.assertEqual(
            'application/octet-stream',
            self.server.request['files']['image'][0]['content_type'])

    def test_upload_content_content_type(self):
        g = self.prepare_form_grab()
        data = b'foo'
        upload_data = UploadContent(data,
                                    content_type='application/grab')
        g.doc.set_input('image', upload_data)
        g.doc.submit(make_request=False)
        post = dict(g.config['multipart_post'])
        self.assertTrue(isinstance(post['image'], UploadContent))

        g.doc.submit()
        self.assertEqual(data,
                         self.server.request['files']['image'][0]['body'])
        self.assertTrue(
            10, len(self.server.request['files']['image'][0]['filename']))
        self.assertEqual(
            'application/grab',
            self.server.request['files']['image'][0]['content_type'])

    # ****************
    # UploadFile Tests
    # ****************

    def test_upload_file(self):
        g = self.prepare_form_grab()
        fh, file_path = tempfile.mkstemp()
        data = b'foo'
        with open(file_path, 'wb') as out:
            out.write(data)
        upload_data = UploadFile(file_path)
        g.doc.set_input('image', upload_data)
        g.doc.submit(make_request=False)
        post = dict(g.config['multipart_post'])
        self.assertTrue(isinstance(post['image'], UploadFile))

        g.doc.submit()
        self.assertEqual(data,
                         self.server.request['files']['image'][0]['body'])
        filename = os.path.split(file_path)[1]
        self.assertEqual(filename,
                         self.server.request['files']['image'][0]['filename'])
        self.assertEqual(
            'application/octet-stream',
            self.server.request['files']['image'][0]['content_type'])

    def test_upload_file_custom_filename(self):
        g = self.prepare_form_grab()
        fh, file_path = tempfile.mkstemp()
        data = b'foo'
        with open(file_path, 'wb') as out:
            out.write(data)
        upload_data = UploadFile(file_path, filename='avatar.jpg')
        g.doc.set_input('image', upload_data)
        g.doc.submit(make_request=False)
        post = dict(g.config['multipart_post'])
        self.assertTrue(isinstance(post['image'], UploadFile))

        g.doc.submit()
        self.assertEqual(data,
                         self.server.request['files']['image'][0]['body'])
        self.assertEqual('avatar.jpg',
                         self.server.request['files']['image'][0]['filename'])
        self.assertEqual(
            'image/jpeg',
            self.server.request['files']['image'][0]['content_type'])

    def test_upload_file_custom_content_type(self):
        g = self.prepare_form_grab()
        fh, file_path = tempfile.mkstemp()
        data = b'foo'
        with open(file_path, 'wb') as out:
            out.write(data)
        upload_data = UploadFile(file_path, filename='avatar.jpg',
                                 content_type='application/grab')
        g.doc.set_input('image', upload_data)
        g.doc.submit(make_request=False)
        post = dict(g.config['multipart_post'])
        self.assertTrue(isinstance(post['image'], UploadFile))

        g.doc.submit()
        self.assertEqual(data,
                         self.server.request['files']['image'][0]['body'])
        self.assertEqual('avatar.jpg',
                         self.server.request['files']['image'][0]['filename'])
        self.assertEqual(
            'application/grab',
            self.server.request['files']['image'][0]['content_type'])



"""
CacheItem interface:
'_id': string,
'url': string,
'response_url': string,
'body': string,
'head': string,
'response_code': int,
'cookies': None,#grab.response.cookies,
"""
from hashlib import sha1
import zlib
import logging
import marshal
import time
from weblib.encoding import make_str

from grab.response import Response
from grab.cookie import CookieManager

logger = logging.getLogger('grab.spider.cache_backend.postgresql')


class CacheBackend(object):
    def __init__(self, database, use_compression=True, spider=None, **kwargs):
        import psycopg2
        from psycopg2.extensions import ISOLATION_LEVEL_READ_COMMITTED

        self.spider = spider
        self.conn = psycopg2.connect(dbname=database, **kwargs)
        self.conn.set_isolation_level(ISOLATION_LEVEL_READ_COMMITTED)
        self.cursor = self.conn.cursor()
        self.cursor.execute("""
            SELECT
                TABLE_NAME
            FROM
                INFORMATION_SCHEMA.TABLES
            WHERE
                TABLE_TYPE = 'BASE TABLE'
            AND
                table_schema NOT IN ('pg_catalog', 'information_schema')""")
        found = False
        for row in self.cursor:
            if row[0] == 'cache':
                found = True
                break
        if not found:
            self.create_cache_table()

    def create_cache_table(self):
        self.cursor.execute('BEGIN')
        self.cursor.execute('''
            CREATE TABLE cache (
                id BYTEA NOT NULL CONSTRAINT primary_key PRIMARY KEY,
                timestamp INT NOT NULL,
                data BYTEA NOT NULL
            );
            CREATE INDEX timestamp_idx ON cache (timestamp);
        ''')
        self.cursor.execute('COMMIT')

    def get_item(self, url, timeout=None):
        """
        Returned item should have specific interface. See module docstring.
        """

        _hash = self.build_hash(url)
        with self.spider.timer.log_time('cache.read.postgresql_query'):
            self.cursor.execute('BEGIN')
            if timeout is None:
                query = ""
            else:
                ts = int(time.time()) - timeout
                query = " AND timestamp > %d" % ts
            sql = '''
                  SELECT data
                  FROM cache
                  WHERE id = %%s %(query)s
                  ''' % {'query': query}
            self.cursor.execute(sql, (_hash,))
            row = self.cursor.fetchone()
            self.cursor.execute('COMMIT')
        if row:
            data = row[0]
            return self.unpack_database_value(data)
        else:
            return None

    def unpack_database_value(self, val):
        with self.spider.timer.log_time('cache.read.unpack_data'):
            dump = zlib.decompress(val)
            return marshal.loads(dump)

    def build_hash(self, url):
        with self.spider.timer.log_time('cache.read.build_hash'):
            utf_url = make_str(url)
            return sha1(utf_url).hexdigest()

    def remove_cache_item(self, url):
        _hash = self.build_hash(url)
        self.cursor.execute('begin')
        self.cursor.execute('''
            DELETE FROM cache WHERE id = %s
        ''', (_hash,))
        self.cursor.execute('commit')

    def load_response(self, grab, cache_item):
        grab.setup_document(cache_item['body'])

        body = cache_item['body']

        def custom_prepare_response_func(transport, g):
            response = Response()
            response.head = cache_item['head']
            response.body = body
            response.code = cache_item['response_code']
            response.download_size = len(body)
            response.upload_size = 0
            response.download_speed = 0
            response.url = cache_item['response_url']
            response.parse()
            response.cookies = CookieManager(transport.extract_cookiejar())
            response.from_cache = True
            return response

        grab.process_request_result(custom_prepare_response_func)

    def save_response(self, url, grab):
        body = grab.response.body

        item = {
            'url': url,
            'response_url': grab.response.url,
            'body': body,
            'head': grab.response.head,
            'response_code': grab.response.code,
            'cookies': None,
        }
        self.set_item(url, item)

    def set_item(self, url, item):
        import psycopg2

        _hash = self.build_hash(url)
        data = self.pack_database_value(item)
        self.cursor.execute('BEGIN')
        ts = int(time.time())
        sql = '''
              UPDATE cache SET timestamp = %s, data = %s WHERE id = %s;
              INSERT INTO cache (id, timestamp, data)
              SELECT %s, %s, %s WHERE NOT EXISTS
                (SELECT 1 FROM cache WHERE id = %s);
              '''
        self.cursor.execute(sql, (ts, psycopg2.Binary(data), _hash,
                            _hash, ts, psycopg2.Binary(data), _hash))
        self.cursor.execute('COMMIT')

    def pack_database_value(self, val):
        dump = marshal.dumps(val)
        return zlib.compress(dump)

    def clear(self):
        self.cursor.execute('BEGIN')
        self.cursor.execute('TRUNCATE cache')
        self.cursor.execute('COMMIT')

    def has_item(self, url, timeout=None):
        """
        Test if required item exists in the cache.
        """

        _hash = self.build_hash(url)
        with self.spider.timer.log_time('cache.read.postgresql_query'):
            if timeout is None:
                query = ""
            else:
                ts = int(time.time()) - timeout
                query = " AND timestamp > %d" % ts
            self.cursor.execute('''
                SELECT id
                FROM cache
                WHERE id = %%s %(query)s
                LIMIT 1
                ''' % {'query': query},
                (_hash,))
            row = self.cursor.fetchone()
        return True if row else False

    def size(self):
        self.cursor.execute('SELECT COUNT(*) from cache')
        return self.cursor.fetchone()[0]



from __future__ import absolute_import
from hashlib import md5
import time
from weblib.encoding import make_str
import os
import mimetypes


class BaseUploadObject(object):
    __slots__ = ()

    def find_content_type(self, filename):
        ctype, encoding = mimetypes.guess_type(filename)
        if ctype is None:
            return 'application/octet-stream'
        else:
            return ctype


class UploadContent(BaseUploadObject):
    __slots__ = ('content', 'filename', 'content_type')

    def __init__(self, content, filename=None, content_type=None):
        self.content = content
        if filename is None:
            self.filename = self.get_random_filename()
        else:
            self.filename = filename
        if content_type is None:
            self.content_type = self.find_content_type(self.filename)
        else:
            self.content_type = content_type

    def get_random_filename(self):
        return md5(make_str(str(time.time()))).hexdigest()[:10]


class UploadFile(BaseUploadObject):
    __slots__ = ('path', 'filename', 'content_type')

    def __init__(self, path, filename=None, content_type=None):
        self.path = path
        if filename is None:
            self.filename = os.path.split(path)[1]
        else:
            self.filename = filename
        if content_type is None:
            self.content_type = self.find_content_type(self.filename)
        else:
            self.content_type = content_type


"""
class UploadContent(str):
    def __new__(cls, value):
        obj = str.__new__(cls, 'xxx')
        obj.raw_value = value
        return obj

    def field_tuple(self):
        # TODO: move to transport extension
        import pycurl
        return pycurl.FORM_CONTENTS, self.raw_value


class UploadFile(str):
    def __new__(cls, path):
        obj = str.__new__(cls, 'xxx')
        obj.path = path
        return obj

    def field_tuple(self):
        # move to transport extension
        import pycurl
        return pycurl.FORM_FILE, self.path
"""



from __future__ import absolute_import
"""
Custom exception which Grab instance could generate.

Taxonomy:

Exception
|-> GrabError
    |-> GrabNetworkError <- IOError
    |-> Grab*Error

Exception
| -> weblib.error.WeblibError
     |-> DataNotFound <- IndexError
"""
import warnings
from weblib.error import DataNotFound  # noqa


class GrabError(Exception):
    """
    All custom Grab exception should be children of that class.
    """


class GrabNetworkError(IOError, GrabError):
    """
    Raises in case of network error.
    """


class GrabTimeoutError(GrabNetworkError):
    """
    Raises when configured time is outed for the request.

    In curl transport it is CURLE_OPERATION_TIMEDOUT (28)
    """


class GrabMisuseError(GrabError):
    """
    Indicates incorrect usage of grab API.
    """


class GrabConnectionError(GrabNetworkError):
    """
    Raised when it is not possible to establish network connection.

    In curl transport it is CURLE_COULDNT_CONNECT (7)
    """


class GrabAuthError(GrabError):
    """
    Raised when remote server denies authentication credentials.

    In curl transport it is CURLE_COULDNT_CONNECT (67)
    """


class GrabTooManyRedirectsError(GrabError):
    """
    Raised when Grab reached max. allowd number of redirects for
    one request.
    """


class GrabDeprecationWarning(Warning):
    """
    Raised when some deprecated feature is used.
    """


class GrabInvalidUrl(GrabError):
    """
    Raised when Grab have no idea how to handle the URL or when
    some error occurred while normalizing URL e.g. IDN processing.
    """


def warn(msg):
    warnings.warn(msg, category=GrabDeprecationWarning, stacklevel=3)



# coding: utf-8
"""
This test case has written to help me
understand how pycurl lib works with cookies
"""
import pycurl
from six import BytesIO
from six.moves.http_cookiejar import CookieJar

from grab.error import GrabMisuseError
from test.util import BaseGrabTestCase
from grab.cookie import create_cookie

# http://xiix.wordpress.com/2006/03/23/mozillafirefox-cookie-format/
# http://curl.haxx.se/libcurl/c/curl_easy_setopt.html
# Cookie:
# * domain
# * whether or not all machines under that domain can read
# the cookie’s information.
# * path
# * Secure Flag: whether or not a secure connection (HTTPS)
# is required to read the cookie.
# * exp. timestamp
# * name
# * value


class TestCookies(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_pycurl_cookies(self):
        self.server.response_once['code'] = 302
        self.server.response_once['cookies'] = {'foo': 'bar', '1': '2'}.items()
        self.server.response_once['headers'] = [
            ('Location', self.server.get_url())]
        self.server.response['get.data'] = 'foo'

        buf = BytesIO()
        header_buf = BytesIO()

        # Configure pycurl instance
        # Usually all these crap is automatically handled by the Grab
        c = pycurl.Curl()
        c.setopt(pycurl.URL, self.server.get_url())
        c.setopt(pycurl.WRITEFUNCTION, buf.write)
        c.setopt(pycurl.HEADERFUNCTION, header_buf.write)
        c.setopt(pycurl.FOLLOWLOCATION, 1)
        c.setopt(pycurl.COOKIEFILE, "")
        c.perform()
        self.assertEqual(b'foo', buf.getvalue())

        print(c.getinfo(pycurl.INFO_COOKIELIST))
        self.assertEquals(2, len(c.getinfo(pycurl.INFO_COOKIELIST)))

        # Just make another request and check that pycurl has
        # submitted two cookies
        c.setopt(pycurl.URL, self.server.get_url())
        c.perform()
        self.assertEquals(2, len(self.server.request['cookies']))

        # Erase cookies
        cookies = c.getinfo(pycurl.INFO_COOKIELIST)
        c.setopt(pycurl.COOKIELIST, "ALL")
        c.setopt(pycurl.URL, self.server.get_url())
        c.perform()
        self.assertEquals(0, len(self.server.request['cookies']))

        # Now let's try to setup pycurl with cookies
        # saved into `cookies` variable
        for cookie in cookies:
            c.setopt(pycurl.COOKIELIST, cookie)
        c.setopt(pycurl.URL, self.server.get_url())
        c.perform()
        self.assertEquals(2, len(self.server.request['cookies']))
        self.assertEquals('bar', self.server.request['cookies']['foo'].value)
        self.assertEquals(set(('foo', '1')),
                          set(self.server.request['cookies'].keys()))

        # Ok, now let's create third cookies that is binded to
        # the path /place, put this cookie into curl object
        # and submit request to /
        # pycurl should not send third cookie
        cookie = '\t'.join((self.server.address,
                            'FALSE', '/place', 'FALSE', '0', 'no', 'way'))
        c.setopt(pycurl.COOKIELIST, cookie)
        c.setopt(pycurl.URL, self.server.get_url())
        c.perform()
        self.assertEquals(set(('foo', '1')),
                          set(self.server.request['cookies'].keys()))

        # Ok, now send request to /place
        c.setopt(pycurl.URL, self.server.get_url('/place'))
        c.perform()
        self.assertEquals(set(('foo', '1', 'no')),
                          set(self.server.request['cookies'].keys()))

        # Now, check that not all cookies set with cookieslist
        # are submitted
        c.setopt(pycurl.COOKIELIST, "ALL")
        c.setopt(pycurl.URL, self.server.get_url())
        c.setopt(pycurl.COOKIELIST, "Set-Cookie: 1=2; domain=microsoft.com")
        c.setopt(pycurl.COOKIELIST, "Set-Cookie: 3=4")
        c.setopt(pycurl.COOKIELIST, "Set-Cookie: 5=6")
        c.perform()
        self.assertEquals(2, len(self.server.request['cookies']))

    def test_cookie(self):
        create_cookie('foo', 'bar', self.server.address)
        self.assertRaises(GrabMisuseError, create_cookie,
                          'foo', 'bar', self.server.address, x='y')

    def test_cookiejar(self):
        c1 = create_cookie('foo', 'bar', self.server.address)
        c2 = create_cookie('foo', 'bar', self.server.address)
        self.assertFalse(c1 == c2)

        c = create_cookie('foo', 'bar', domain='.dumpz.org')
        self.assertEquals(c.domain, '.dumpz.org')

        cj = CookieJar()
        cj.set_cookie(create_cookie('foo', 'bar', domain='foo.com'))
        cj.set_cookie(create_cookie('foo', 'bar', domain='bar.com'))
        self.assertEqual(len(cj), 2)



import six
from grab.spider import Spider, Task
from grab.spider.error import SpiderMisuseError
from unittest import TestCase
from grab.spider.queue_backend.base import QueueInterface

from test.util import BaseGrabTestCase, build_spider
from test_settings import MONGODB_CONNECTION, REDIS_CONNECTION


class SpiderQueueMixin(object):
    class SimpleSpider(Spider):
        def task_page(self, grab, task):
            self.stat.collect('url_history', task.url)
            self.stat.collect('priority_history', task.priority)

    def test_basic_priority(self):
        bot = build_spider(self.SimpleSpider, parser_pool_size=1,
                           thread_number=1)
        self.setup_queue(bot)
        bot.taskq.clear()
        requested_urls = {}
        for priority in (4, 2, 1, 5):
            url = self.server.get_url() + '?p=%d' % priority
            requested_urls[priority] = url
            bot.add_task(Task('page', url=url,
                              priority=priority))
        bot.run()
        urls = [x[1] for x in sorted(requested_urls.items(),
                                     key=lambda x: x[0])]
        self.assertEqual(urls, bot.stat.collections['url_history'])

    def test_queue_length(self):
        bot = build_spider(self.SimpleSpider)
        self.setup_queue(bot)
        bot.taskq.clear()
        for x in six.moves.range(5):
            bot.add_task(Task('page', url=self.server.get_url()))
        self.assertEqual(5, bot.taskq.size())
        bot.run()
        self.assertEqual(0, bot.taskq.size())
        bot.run()

    def test_taskq_render_stats(self):
        bot = build_spider(self.SimpleSpider)
        bot.render_stats()


class SpiderMemoryQueueTestCase(BaseGrabTestCase, SpiderQueueMixin):
    def setup_queue(self, bot):
        bot.setup_queue(backend='memory')

    def test_schedule(self):
        """
        In this test I create a number of delayed task
        and then check the order in which they was executed
        """
        server = self.server

        class TestSpider(Spider):
            def task_generator(self):
                yield Task('page', url=server.get_url(), delay=1.5, num=3)
                yield Task('page', url=server.get_url(), delay=4.5, num=2)
                yield Task('page', url=server.get_url(), delay=3, num=4)
                yield Task('page', url=server.get_url(), num=1)

            def task_page(self, grab, task):
                self.stat.collect('numbers', task.num)

        bot = build_spider(TestSpider, thread_number=1)
        self.setup_queue(bot)
        bot.run()
        self.assertEqual(bot.stat.collections['numbers'], [1, 3, 4, 2])


class BasicSpiderTestCase(SpiderQueueMixin, BaseGrabTestCase):
    _backend = 'mongo'

    def setup_queue(self, bot):
        bot.setup_queue(backend='mongo', **MONGODB_CONNECTION)

    def test_schedule(self):
        """
        In this test I create a number of delayed task
        and then check the order in which they was executed
        """
        server = self.server

        class TestSpider(Spider):
            def task_generator(self):
                yield Task('page', url=server.get_url(), num=1)
                yield Task('page', url=server.get_url(), delay=1.5, num=2)
                yield Task('page', url=server.get_url(), delay=0.5, num=3)
                yield Task('page', url=server.get_url(), delay=1, num=4)

            def task_page(self, grab, task):
                self.stat.collect('numbers', task.num)

        bot = build_spider(TestSpider)
        self.setup_queue(bot)
        bot.run()
        self.assertEqual(bot.stat.collections['numbers'], [1, 3, 4, 2])

    def test_clear_collection(self):
        bot = build_spider(self.SimpleSpider)
        self.setup_queue(bot)
        bot.taskq.clear()


class SpiderRedisQueueTestCase(SpiderQueueMixin, BaseGrabTestCase):
    _backend = 'redis'

    def setup_queue(self, bot):
        bot.setup_queue(backend='redis', **REDIS_CONNECTION)

    def test_delay_error(self):
        bot = build_spider(self.SimpleSpider)
        self.setup_queue(bot)
        bot.taskq.clear()
        self.assertRaises(SpiderMisuseError,
                          bot.add_task,
                          Task('page', url=self.server.get_url(), delay=1))


class QueueInterfaceTestCase(TestCase):
    def test_abstract_methods(self):
        """Just to improve test coverage"""
        class BrokenQueue(QueueInterface):
            pass

        taskq = BrokenQueue('spider_name')
        self.assertRaises(NotImplementedError, taskq.put, None, None)
        self.assertRaises(NotImplementedError, taskq.get)
        self.assertRaises(NotImplementedError, taskq.size)
        self.assertRaises(NotImplementedError, taskq.clear)



import logging
import os
from argparse import ArgumentParser
import six

from grab.util.config import build_spider_config, build_root_config
from grab.util.module import load_spider_class
from weblib.logs import default_logging
from weblib.files import clear_directory
from weblib.encoding import make_str

logger = logging.getLogger('grab.script.crawl')


def setup_arg_parser(parser):
    parser.add_argument('spider_name', type=str)
    parser.add_argument('-t', '--thread-number', default=None, type=int,
                        help='Number of network threads')
    parser.add_argument('--slave', action='store_true', default=False,
                        help='Enable the slave-mode')
    parser.add_argument('-n', '--network-logs', action='store_true',
                        default=False,
                        help='Dump to console details about network requests')
    parser.add_argument('--disable-proxy', action='store_true', default=False,
                        help='Disable proxy servers')
    parser.add_argument('--ignore-lock', action='store_true', default=False)
    parser.add_argument('--disable-report', action='store_true', default=False)
    parser.add_argument('--disable-default-logs', action='store_true',
                        default=False)
    parser.add_argument('--settings-module', type=str, default='settings')
    parser.add_argument('--api-port', type=int, default=None)
    parser.add_argument('--multiprocess', action='store_true', default=False,
                        help='Run task handlers (HTML parsers) in separate '
                             'processes')
    parser.add_argument('--parser-pool-size', type=int)


def get_lock_key(spider_name, lock_key=None, ignore_lock=False,
                 slave=False, **kwargs):
    # --ignore-lock has highest precedence
    if ignore_lock:
        return None

    # If --lock-key is specified explicitly use it
    if lock_key is not None:
        return lock_key

    # Do not lock --slave spiders
    if slave:
        return None

    # As fallback, if no information has been given about locking
    # generate lock key from the spider name and use it
    lock_key = 'crawl.%s' % spider_name
    return lock_key


def save_list(lst, path):
    """
    Save items from list to the file.
    """

    with open(path, 'wb') as out:
        lines = []
        for item in lst:
            if isinstance(item, (six.text_type, six.binary_type)):
                lines.append(make_str(item))
            else:
                lines.append(make_str(json.dumps(item)))
        out.write(b'\n'.join(lines) + b'\n')


def main(spider_name, thread_number=None, slave=False,
         settings_module='settings', network_logs=False,
         disable_proxy=False, ignore_lock=False,
         disable_report=False,
         disable_default_logs=False,
         api_port=None,
         multiprocess=False,
         parser_pool_size=None,
         *args, **kwargs):
    if disable_default_logs:
        default_logging(propagate_network_logger=network_logs,
                        grab_log=None, network_log=None)
    else:
        default_logging(propagate_network_logger=network_logs)

    root_config = build_root_config(settings_module)
    spider_class = load_spider_class(root_config, spider_name)
    spider_config = build_spider_config(spider_class, root_config)

    spider_args = None
    if hasattr(spider_class, 'setup_arg_parser'):
        parser = ArgumentParser()
        spider_class.setup_arg_parser(parser)
        opts, trash = parser.parse_known_args()
        spider_args = vars(opts)

    bot = spider_class(
        thread_number=thread_number,
        slave=slave,
        config=spider_config,
        network_try_limit=None,
        task_try_limit=None,
        args=spider_args,
        http_api_port=api_port,
        multiprocess=multiprocess,
        parser_pool_size=parser_pool_size,
    )
    opt_queue = spider_config.get('queue')
    if opt_queue:
        bot.setup_queue(**opt_queue)

    opt_cache = spider_config.get('cache')
    if opt_cache:
        bot.setup_cache(**opt_cache)

    opt_proxy_list = spider_config.get('proxy_list')
    if opt_proxy_list:
        if disable_proxy:
            logger.debug('Proxy servers disabled via command line')
        else:
            bot.load_proxylist(**opt_proxy_list)

    opt_ifaces = spider_config.get('command_interfaces')
    if opt_ifaces:
        for iface_config in opt_ifaces:
            bot.controller.add_interface(**iface_config)

    try:
        bot.run()
    except KeyboardInterrupt:
        pass

    stats = bot.render_stats(timing=spider_config.get('display_timing'))
    stats_with_time = bot.render_stats(timing=True)

    if spider_config.get('display_stats'):
        logger.debug(stats)

    pid = os.getpid()
    logger.debug('Spider pid is %d' % pid)

    if not disable_report:
        if spider_config.get('save_report'):
            for subdir in (str(pid), 'last'):
                dir_ = 'var/%s' % subdir
                if not os.path.exists(dir_):
                    os.makedirs(dir_)
                else:
                    clear_directory(dir_)
                for key, lst in bot.items.items():
                    fname_key = key.replace('-', '_')
                    save_list(lst, '%s/%s.txt' % (dir_, fname_key))
                with open('%s/report.txt' % dir_, 'wb') as out:
                    out.write(make_str(stats_with_time))

    return {
        'spider_stats': bot.render_stats(timing=False),
        'spider_timing': bot.render_timing(),
    }



"""
RTFM:

* http://docs.python.org/2/library/cookielib.html#cookie-objects

Some code got from
    https://github.com/kennethreitz/requests/blob/master/requests/cookies.py
"""
from __future__ import absolute_import
from six.moves.http_cookiejar import CookieJar, Cookie
import json
import logging

from grab.error import GrabMisuseError

logger = logging.getLogger('grab.cookie')
COOKIE_ATTRS = ('name', 'value', 'version', 'port', 'domain',
                'path', 'secure', 'expires', 'discard', 'comment',
                'comment_url', 'rfc2109')


def create_cookie(name, value, domain, httponly=None, **kwargs):
    "Creates `cookielib.Cookie` instance"

    if domain == 'localhost':
        domain = ''
    config = dict(
        name=name,
        value=value,
        version=0,
        port=None,
        domain=domain,
        path='/',
        secure=False,
        expires=None,
        discard=True,
        comment=None,
        comment_url=None,
        rfc2109=False,
        rest={'HttpOnly': httponly},
    )

    for key in kwargs.keys():
        if key not in config:
            raise GrabMisuseError('Function `create_cookie` does not accept '
                                  '`%s` argument' % key)

    config.update(**kwargs)
    config['rest']['HttpOnly'] = httponly

    config['port_specified'] = bool(config['port'])
    config['domain_specified'] = bool(config['domain'])
    config['domain_initial_dot'] = (config['domain'] or '').startswith('.')
    config['path_specified'] = bool(config['path'])

    return Cookie(**config)


class CookieManager(object):
    """
    Each Grab instance has `cookies` attribute that is instance of
    `CookieManager` class.

    That class contains helpful methods to create, load, save cookies from/to
    different places.
    """

    __slots__ = ('cookiejar',)

    def __init__(self, cookiejar=None):
        if cookiejar is not None:
            self.cookiejar = cookiejar
        else:
            self.cookiejar = CookieJar()
        # self.disable_cookiejar_lock(self.cookiejar)

    # def disable_cookiejar_lock(self, cj):
        # cj._cookies_lock = dummy_threading.RLock()

    def set(self, name, value, domain, **kwargs):
        """Add new cookie or replace existing cookie with same parameters.

        :param name: name of cookie
        :param value: value of cookie
        :param kwargs: extra attributes of cookie
        """

        if domain == 'localhost':
            domain = ''

        self.cookiejar.set_cookie(create_cookie(name, value, domain, **kwargs))

    def update(self, cookies):
        if isinstance(cookies, CookieJar):
            for cookie in cookies:
                self.cookiejar.set_cookie(cookie)
        elif isinstance(cookies, CookieManager):
            for cookie in cookies.cookiejar:
                self.cookiejar.set_cookie(cookie)
        else:
            raise GrabMisuseError('Unknown type of cookies argument: %s'
                                  % type(cookies))

    @classmethod
    def from_cookie_list(cls, clist):
        cj = CookieJar()
        for cookie in clist:
            cj.set_cookie(cookie)
        return cls(cj)

    def clear(self):
        self.cookiejar = CookieJar()

    def __getstate__(self):
        state = {}
        for cls in type(self).mro():
            cls_slots = getattr(cls, '__slots__', ())
            for slot in cls_slots:
                if slot != '__weakref__':
                    if hasattr(self, slot):
                        state[slot] = getattr(self, slot)

        state['_cookiejar_cookies'] = list(self.cookiejar)
        del state['cookiejar']

        return state

    def __setstate__(self, state):
        state['cookiejar'] = CookieJar()
        for cookie in state['_cookiejar_cookies']:
            state['cookiejar'].set_cookie(cookie)
        del state['_cookiejar_cookies']

        for slot, value in state.items():
            setattr(self, slot, value)

    def __getitem__(self, key):
        for cookie in self.cookiejar:
            if cookie.name == key:
                return cookie.value
        raise KeyError

    def items(self):
        res = []
        for cookie in self.cookiejar:
            res.append((cookie.name, cookie.value))
        return res

    def load_from_file(self, path):
        """
        Load cookies from the file.

        Content of file should be a JSON-serialized list of dicts.
        """

        with open(path) as inf:
            data = inf.read()
            if data:
                items = json.loads(data)
            else:
                items = {}
        for item in items:
            extra = dict((x, y) for x, y in item.items()
                         if x not in ['name', 'value', 'domain'])
            self.set(item['name'], item['value'], item['domain'], **extra)

    def get_dict(self):
        res = []
        for cookie in self.cookiejar:
            res.append(dict((x, getattr(cookie, x)) for x in COOKIE_ATTRS))
        return res

    def save_to_file(self, path):
        """
        Dump all cookies to file.

        Cookies are dumped as JSON-serialized dict of keys and values.
        """

        with open(path, 'w') as out:
            out.write(json.dumps(self.get_dict()))



# coding: utf-8
from test.util import build_grab
from test.util import BaseGrabTestCase


class GrabSimpleTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_empty_useragent_pycurl(self):
        g = build_grab()

        # Empty string disable default pycurl user-agent
        g.setup(user_agent='')
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['headers']
                             .get('user-agent', ''), '')

    def test_useragent_simple(self):
        g = build_grab()

        # Simple case: setup user agent manually
        g.setup(user_agent='foo')
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['headers']['user-agent'], 'foo')

    def test_useragent(self):
        g = build_grab()

        # Null value activates default random user-agent
        # For some transports it just allow them to send default user-agent
        # like in Kit transport case
        g = build_grab()
        g.setup(user_agent=None)
        g.go(self.server.get_url())
        self.assertTrue(len(self.server.request['headers']) > 0)
        self.assertFalse('PycURL' in
                         self.server.request['headers']['user-agent'])

        # By default user_agent is None => random user agent is generated
        g = build_grab()
        g.go(self.server.get_url())
        self.assertTrue(len(self.server.request['headers']) > 0)
        self.assertFalse('PycURL' in
                         self.server.request['headers']['user-agent'])

        # Simple case: setup user agent manually
        g.setup(user_agent='foo')
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['headers']['user-agent'], 'foo')

        # user agent from file should be loaded
        path = '/tmp/__ua.txt'
        open(path, 'w').write('GOD')
        g.setup(user_agent=None, user_agent_file=path)
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['headers']['user-agent'], 'GOD')

        # random user agent from file should be loaded
        path = '/tmp/__ua.txt'
        open(path, 'w').write('GOD1\nGOD2')
        g.setup(user_agent=None, user_agent_file=path)
        g.go(self.server.get_url())
        self.assertTrue(self.server.request['headers']['user-agent']
                        in ('GOD1', 'GOD2'))
        ua = g.config['user_agent']

        # User-agent should not change
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['headers']['user-agent'], ua)

        # User-agent should not change
        g.go(self.server.get_url())
        self.assertEqual(self.server.request['headers']['user-agent'], ua)



"""
This is obsoleted module. Use selection package instead.
"""

from selection.selector_list import SelectorList, RexResultList  # noqa
from selection.backend.xpath import XpathSelector as XpathSelectorOrigin
from selection.backend.pyquery import PyquerySelector  # noqa
from grab.error import warn


class XpathSelector(XpathSelectorOrigin):
    def __init__(self, *args, **kwargs):
        warn('You are using XpathSelector from deprecated `grab.selector` '
             'package. Please, switch to `selection` package.')
        super(XpathSelector, self).__init__(*args, **kwargs)



from test.util import BaseGrabTestCase
from test.util import build_grab


class ExtensionPyqueryTestCase(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_pyquery_handler(self):
        self.server.response['get.data'] =\
            '<body><h1>Hello world</h1><footer>2014</footer>'
        g = build_grab()
        g.go(self.server.get_url())

        self.assertEqual(g.doc.pyquery('h1').text(), 'Hello world')



from six.moves.SimpleHTTPServer import SimpleHTTPRequestHandler
from six.moves.socketserver import TCPServer
import logging
import os
import threading
import json
from weblib.encoding import make_str

logger = logging.getLogger('grab.spider.http_api')
BASE_DIR = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))


class ApiHandler(SimpleHTTPRequestHandler):
    def do_GET(self):
        if self.path == '/':
            self.home()
        elif self.path == '/api/info':
            self.api_info()
        elif self.path == '/api/stop':
            self.api_stop()
        else:
            self.not_found()

    def response(self, code=200, content=b'',
                 content_type='text/html; charset=utf-8'):
        self.send_response(code)
        self.send_header('Content-type', content_type)
        self.end_headers()
        self.wfile.write(content)

    def not_found(self):
        self.response(404)

    def api_info(self):
        info = {
            'counters': self.spider.stat.counters,
            'collections': dict((x, len(y)) for (x, y)
                                in self.spider.stat.collections.items()),
            'thread_number': self.spider.thread_number,
            'parser_pool_size': self.spider.parser_pool_size,
        }
        content = make_str(json.dumps(info))
        self.response(content=content)

    def api_stop(self):
        self.response()
        self.spider.stop()

    def home(self):
        html_file = os.path.join(BASE_DIR, 'spider/static/http_api.html')
        content = open(html_file, 'rb').read()
        self.response(content=content)


class ReuseTCPServer(TCPServer):
    allow_reuse_address = True


class HttpApiThread(threading.Thread):
    def __init__(self, spider, *args, **kwargs):
        self.spider = spider
        super(HttpApiThread, self).__init__(*args, **kwargs)

    def run(self):
        ApiHandler.spider = self.spider
        self.server = ReuseTCPServer(("", self.spider.http_api_port),
                                     ApiHandler)
        logging.debug('Serving HTTP API on localhost:%d'
                      % self.spider.http_api_port)
        self.server.serve_forever()



from unittest import TestCase

from test.mixin.spider_cache import SpiderCacheMixin


class BasicSpiderTestCase(TestCase, SpiderCacheMixin):
    def setUp(self):
        SpiderCacheMixin.setUp(self)

    def setup_cache(self, bot):
        bot.setup_cache(backend='mongo', database='test_spider')



# coding: utf-8
try:
    import cPickle as pickle
except ImportError:
    import pickle
from multiprocessing import Queue

from test.util import BaseGrabTestCase
from test.util import build_grab


class TestGrab(BaseGrabTestCase):
    def setUp(self):
        self.server.reset()

    def test_pickling(self):
        """
        Test that Grab instance could be pickled and unpickled.
        """

        g = build_grab()
        self.server.response['get.data'] =\
            '<form><textarea name="text">the cat</textarea></form>'
        g.go(self.server.get_url())
        g.set_input('text', 'foobar')
        data = pickle.dumps(g, pickle.HIGHEST_PROTOCOL)

        def func(pickled_grab, resultq):
            g2 = pickle.loads(pickled_grab)
            text = g2.doc.select('//textarea').text()
            resultq.put(text)

        result_queue = Queue()
        # p = Process(target=func, args=[data, result_queue])
        # p.start()
        func(data, result_queue)

        text = result_queue.get(block=True, timeout=1)
        self.assertEqual(text, 'the cat')



from unittest import TestCase
from grab.util.module import (build_spider_registry, SPIDER_REGISTRY,
                              load_spider_class)
from grab.spider import Spider
from grab.spider.error import SpiderInternalError
from test.util import TMP_DIR
import os
import sys

class BaseSpider(Spider):
    class Meta:
        abstract = True


class FirstSpider(BaseSpider):
    pass


class SecondSpider(BaseSpider):
    pass


class UtilModuleTestCase(TestCase):
    def build_config(self, modules):
        cfg = {'global': {'spider_modules': modules}}
        return cfg

    def test_build_spider_registry(self):
        cfg = self.build_config(['test.util_module', 'zz'])
        reg = build_spider_registry(cfg)
        self.assertEqual(reg, SPIDER_REGISTRY)

    def test_build_spider_registry_with_name(self):
        cfg = self.build_config(['test.util_module', 'zz'])
        reg = build_spider_registry(cfg)
        self.assertEqual(2, len(reg))

        cfg = self.build_config(['test.util_module:SecondSpider', 'zz'])
        reg = build_spider_registry(cfg)
        self.assertEqual(1, len(reg))

    def test_build_spider_registry_failed_module(self):
        cfg = self.build_config(['test.util_module', 'zz',
                                 'test.files.invalid_import'])
        reg = build_spider_registry(cfg)
        self.assertEqual(2, len(reg))

    def test_build_spider_registry_same_name_spiders(self):
        cfg = self.build_config(['test.util_module', 'zz',
                                 'test.files.first_spider'])
        self.assertRaises(SpiderInternalError, build_spider_registry, cfg)

    def test_load_spider_class(self):
        cfg = self.build_config(['test.util_module'])
        SPIDER_REGISTRY.clear()
        cls = load_spider_class(cfg, 'first_spider')
        self.assertEqual(cls, FirstSpider)

    def test_load_spider_class_error(self):
        cfg = self.build_config(['test.util_module'])
        reg = build_spider_registry(cfg)
        self.assertRaises(SpiderInternalError, load_spider_class,
                          cfg, 'first_spider_zzz')
